/*
 * differential timing analysis simulation.
 *
 * Copyright (c) 2003 by bennet yee <bsy@cs.ucsd.edu>, all rights reserved.
 * see dta.h for more information.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>

#include "dta.h"
#include "dta_priv.h"

/*
 * use a good random hash function to create products; doing something
 * ad-hoc can result in having accidental correlations from cycles etc
 * that appear to be a signal, but isn't.
 */
#include <openssl/md5.h>

/*
 * model of a differential timing analysis vulnerable system.
 *
 * we have a black-box timing model, which is used in constructing
 * the simulated smartcard.
 */

/*
 * timing cost model.  all we need is a random function of the inputs
 * that has reasonable distribution.
 *
 * units might be cycles, microseconds, etc.
 */
static sim_time_t	cost_model(big_int_t const	*x,
				   big_int_t const	*y,
				   big_int_t const	*n)
{
	sim_time_t	t;

	t = ((x->val[0] ^ x->val[1]) + (y->val[0] * y->val[1]) + (y->val[2] & x->val[3]) + (x->val[2] - y->val[3])) % 601;
	return t + 100;
}

void	exp_canonicalize(exp_t	*ep)
{
	int	len = ep->length;
	while (--len >= 0 && ep->bits[len] == 0)
		;
	ep->length = len + 1;
}

/* assumes canonicalized */
int	exp_equal(exp_t const	*a,
		  exp_t const	*b)
{
	int	i;

	if ((i = a->length) != b->length) return 0;
	while (--i >= 0) if (a->bits[i] != b->bits[i]) return 0;
	return 1;
}

/*
 * initialize a big integer from a small one.
 */
void	big_int_init(big_int_t	*var,
		     uint32_t	val)
{
	memset(var->val,0,sizeof var->val);
	var->val[0] = val;
}

int	big_int_equal(big_int_t const	*a,
		      big_int_t	const	*b)
{
	int	i;

	for (i = NLIMBS; --i >= 0;) {
		if (a->val[i] != b->val[i]) return 0;
	}
	return 1;
}

void	big_int_print(big_int_t const	*var,
		      FILE		*iob)
{
	int	i, sep = '[';

	for (i = 0; i < NLIMBS; i++) {
		fprintf(iob,"%c%08x",sep,var->val[i]);
		sep = ':';
	}
	putchar(']');
}

/*
 * timing model for modular multiply.  we do the operation and return
 * both the result and the time used to compute it.
 */
void	modmult(big_int_t	*prod,
		sim_time_t	*cost,
		big_int_t const	*x,
		big_int_t const	*y,
		big_int_t const	*n)
{
	MD5_CTX		ctx;

	*cost = cost_model(x,y,n); /* compute cost from original x,y */
	/*
	 * "result" has only to be a deterministic random appearing function
	 * of the inputs
	 */
	MD5_Init(&ctx);
	MD5_Update(&ctx,x->val,sizeof x->val);
	MD5_Update(&ctx,y->val,sizeof y->val);
	/* prod may be an alias for x or y */
	MD5_Final((unsigned char *) prod->val,&ctx);
}

/*
 * generate raw data for external analysis.  outputs timing information
 * for each stage; with negated value when exp bit zero and mod mult is
 * not actually performed.
 */
void	modexp_data(big_int_t const	*x,
		    exp_t const		*e,
		    big_int_t const	*n)
{
	big_int_t	y, z, w;
	int		i;
	sim_time_t	t;

	big_int_init(&y,1); z = *x;
	big_int_print(x,stdout);

	for (i = 0; i < e->length; i++) {
		if (0 != e->bits[i]) {
			modmult(&y,&t, &y,&z,n); /* y <- y * z mod n */
			printf(" %4d",t);
		} else {
			modmult(&w,&t, &y,&z,n); /* w <- y * z mod n */
			printf(" %4d",t);
		}
		modmult(&z,&t, &z,&z,n); /* z <- z * z mod n */
	}
}

/*
 * modular exponentiation.  w <- x^e mod n (simulated),
 *  z <- x^2^e.length mod n (simulated), and
 *  tot <- time used to do the computation, with the z
 *  compute time discounted.
 */
void	modexp_z(big_int_t		*w,
		 big_int_t 		*zout,
		 sim_time_t		*tot,
		 big_int_t const	*x,
		 exp_t const		*e,
		 big_int_t const	*n)
{
	big_int_t	y, z, discard;
	int		i;
	sim_time_t	sofar = 0, t;

	big_int_init(&y,1); z = *x;

	for (i = 0; i < e->length; i++) {
		if (0 != e->bits[i]) {
			modmult(&y,&t, &y,&z,n); /* y <- y * z mod n */
			sofar += t;
		} else {
			modmult(&discard,&t, &y,&z,n);
			sofar += t;
		}
		modmult(&z,&t, &z,&z,n); /* z <- z * z mod n */
		/* sofar += t; discount the time for z computation */
	}
	*w = y;
	*tot = sofar;
	*zout = z;
}

void	modexp(big_int_t	*w,
	       sim_time_t	*tot,
	       big_int_t const	*x,
	       exp_t const	*e,
	       big_int_t const	*n)
{
	big_int_t	zout;

	modexp_z(w,&zout,tot, x,e,n);
}

void	bigrand(big_int_t	*r)
{
	int	i;

	for (i = 0; i < NLIMBS; i++)
		r->val[i] = random();
}

void	print_exp(exp_t const	*e,
		  FILE		*iob)
{
	int	i;

	fprintf(iob,"Exponent, low-order bit first:\n");
	for (i = 0; i < e->length; i++) {
		if ((i & 0x1f) == 0) putc('\n',iob);
		else if ((i & 0xf) == 0) fputs("  ",iob);
		else if ((i & 0x3) == 0) putc(' ',iob);
		putc(e->bits[i] ? '1' : '0', iob);
	}
	putc('\n',iob);
}

int	load_exp(exp_t	*e,
		 FILE	*iob)
{
	int	i, ch;

	for (i = 0; i < MAXBITS; i++) {
again:		switch (ch = getc(iob)) {
		case ' ':
		case '\n':
			goto again;
		case EOF:
			goto done;
		case '0':
			e->bits[i] = 0;
			break;
		case '1':
			e->bits[i] = 1;
			break;
		default:
			return 0;
		}
	}
done:
	e->length = i;
	return 1;
}
